
#include <stdio.h>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdlib.h>
#include <ctype.h>
#include "gloutput.h"
#include "vsync.h"

static SDL_Surface *surface;
static int output_width, output_height;

static GLuint texture;
static GLubyte *pixels;
static int tex_width, tex_height;

static void gl_enter_2d_mode( void )
{
    glDisable( GL_DEPTH_TEST );
    glDisable( GL_CULL_FACE );

    glDisable( GL_BLEND );
    // glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

    glViewport( 0, 0, output_width, output_height );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

    glOrtho( 0.0, output_width, output_height, 0.0, -1.0, 1.0 );
    // gluOrtho2D( 0.0, 1.0, 0.0, 1.0 );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
}


/* Quick utility function for texture creation */
static int power_of_two( int input )
{
    int value = 1;
    while( value < input ) value <<= 1;
    return value;
}

int gloutput_init( int width, int height )
{
    if( SDL_Init( SDL_INIT_VIDEO ) < 0 ) return 0;

    if( !vsync_init() ) return 0;

    SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 1 );
    SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 1 );
    SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 1 );
    SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 0 );

    surface = SDL_SetVideoMode( width, height, 32, SDL_OPENGL );
    if( !surface ) {
        fprintf( stderr, "no surface.\n" );
        return 0;
    }
    SDL_ShowCursor( SDL_DISABLE );
    glViewport( 0, 0, width, height);
    output_width = width;
    output_height = height;

    /* Stupid power-of-two textures. */
    tex_width = power_of_two( width );
    tex_height = power_of_two( height );
    fprintf( stderr, "tex width %d, tex height %d\n", tex_width, tex_height );

    glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );
    glGenTextures( 1, &texture );    
    pixels = malloc( tex_width * tex_height * 4 );

    glDisable( GL_DEPTH_TEST );
    glDisable( GL_CULL_FACE );
    glDisable( GL_LIGHTING );
    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );

    gl_enter_2d_mode();

    glEnable( GL_TEXTURE_2D );
    glBindTexture( GL_TEXTURE_2D, texture );
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, tex_width, tex_height, 0,
                  GL_RGBA, GL_UNSIGNED_BYTE, 0 );

    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    // glTexEnvi       (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,   GL_REPLACE);

    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
    glClear( GL_COLOR_BUFFER_BIT );
    glDisable( GL_TEXTURE_2D );

    glFlush();

    return 1;
}

void gloutput_shutdown( void )
{
    SDL_Quit();
    vsync_shutdown();
}

void gloutput_lock_output_buffer( void )
{
}

unsigned char *gloutput_get_output_buffer( void )
{
    return pixels;
}

int gloutput_get_output_stride( void )
{
    return output_width*4;
}

void gloutput_unlock_output_buffer( void )
{
}

void gloutput_show_frame( void )
{
    double wtex = (double) output_width / (double) tex_width;
    double htex = (double) output_height / (double) tex_height;

    glClear( GL_COLOR_BUFFER_BIT );
    glEnable( GL_TEXTURE_2D );
    glBindTexture( GL_TEXTURE_2D, texture );
    glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, output_width, output_height, GL_RGBA, GL_UNSIGNED_BYTE, pixels );

    // glTexEnvi       (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,   GL_REPLACE);

    // glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, tex_width, tex_height, 0,
    //               GL_RGBA, GL_UNSIGNED_BYTE, pixels );

    // glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    // glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

    glBegin( GL_QUADS );
    glTexCoord2f( 0.0f, htex ); glVertex2f( 0.0f, 0.0f );
    glTexCoord2f( 0.0f, 0.0f ); glVertex2f( 0.0f, 1.0f );
    glTexCoord2f( wtex, 0.0f ); glVertex2f( 1.0f, 1.0f );
    glTexCoord2f( wtex, htex ); glVertex2f( 1.0f, 0.0f );
    glEnd();
    glDisable( GL_TEXTURE_2D );

/*
    glPixelZoom( 1.0, -1.0 );
    glRasterPos2i( 0, 1 );
    glDrawPixels( output_width, output_height, GL_RGBA, GL_UNSIGNED_BYTE, pixels );
*/

    glFlush();
    // SDL_GL_SwapBuffers();
}

void gloutput_wait_for_retrace( void )
{
    vsync_wait_for_retrace();
}

void gloutput_set_window_caption( const char *cpation )
{
}

void gloutput_poll_events( input_t *in )
{
    SDL_Event event;
    int curcommand = 0, arg = 0;
    SDLMod mods;

    while( SDL_PollEvent( &event ) ) {

        if( event.type == SDL_QUIT ) {
            curcommand = I_QUIT;
            input_callback( in, curcommand, arg );
        }

        if( event.type == SDL_KEYDOWN ) {
            curcommand = I_KEYDOWN;

            mods = event.key.keysym.mod;
            if( mods & KMOD_SHIFT ) arg |= I_SHIFT;
            if( mods & KMOD_CTRL ) arg |= I_CTRL;
            if( mods & (KMOD_META | KMOD_ALT) ) arg |= I_META;

            switch ( event.key.keysym.sym ) {

            case SDLK_KP_PLUS:     arg |= '+'; break;
            case SDLK_KP_MINUS:    arg |= '-'; break;
            case SDLK_KP_DIVIDE:   arg |= '/'; break;
            case SDLK_KP_PERIOD:   arg |= '.'; break;
            case SDLK_KP_MULTIPLY: arg |= '*'; break;
            case SDLK_KP_EQUALS:   arg |= '='; break;

            case SDLK_KP0: arg |= '0'; break;
            case SDLK_KP1: arg |= '1'; break;
            case SDLK_KP2: arg |= '2'; break;
            case SDLK_KP3: arg |= '3'; break;
            case SDLK_KP4: arg |= '4'; break;
            case SDLK_KP5: arg |= '5'; break;
            case SDLK_KP6: arg |= '6'; break;
            case SDLK_KP7: arg |= '7'; break;
            case SDLK_KP8: arg |= '8'; break;
            case SDLK_KP9: arg |= '9'; break;

            case SDLK_KP_ENTER:
                arg |= I_ENTER;
                break;

            default:
                if( (mods & KMOD_SHIFT) || (mods & KMOD_CAPS) ) {
                    arg |= toupper( event.key.keysym.sym );
                } else {
                    arg |= event.key.keysym.sym;
                }
                break;
            }

            input_callback( in, curcommand, arg );
        }

        if( event.type == SDL_MOUSEBUTTONDOWN ) {

            input_callback( in, I_BUTTONPRESS, event.button.button );

        }
    }
}

output_api_t gloutput_api = {
    gloutput_init,
    gloutput_shutdown,

    gloutput_lock_output_buffer,
    gloutput_get_output_buffer,
    gloutput_get_output_stride,
    gloutput_unlock_output_buffer,

    gloutput_show_frame,
    gloutput_wait_for_retrace,
    gloutput_set_window_caption,
    gloutput_poll_events
};

output_api_t *gloutput_get_output( void )
{
    return &gloutput_api;
}

