

#include <stdio.h>
#include <sys/time.h>
#include "vsync.h"

static int timediff( struct timeval *large, struct timeval *small )
{
    return (   ( ( large->tv_sec * 1000 * 1000 ) + large->tv_usec )
             - ( ( small->tv_sec * 1000 * 1000 ) + small->tv_usec ) );
}

int main( int argc, char **argv )
{
    int trials = 100;
    int sum = 0;
    int i;

    if( !vsync_init() ) {
        fprintf( stderr, "testsync: Can't open DRI device, exiting.\n" );
    }

    vsync_wait_for_retrace();
    vsync_wait_for_retrace();
    for( i = 0; i < trials; i++ ) {
        struct timeval bef, af;
        gettimeofday( &bef, 0 );
        vsync_wait_for_retrace();
        gettimeofday( &af, 0 );
        sum += timediff( &af, &bef );
    }
    fprintf( stderr, "testsync: Refresh roughly measured at %d usec over %d trials.\n",
             sum / trials, trials );
    fprintf( stderr, "testsync: Estimated refresh rate: %.4fhz.\n",
             (1000.0 * 1000.0) / ((double) (sum / trials)) );
    vsync_shutdown();
    return 1;
}

