
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/xf86vmode.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include "tools.h"
#include "xoutput.h"
#include "vsync.h"

const int use_xshm_directly = 0;

static Atom wmProtocolsAtom;
static Atom wmDeleteAtom;
static char *atomNames[] = { "WM_PROTOCOLS", "WM_DELETE_WINDOW" };

extern int XShmGetEventBase(Display *dpy);
static int CompletionType;

enum types {
    MODE_RGB_32,
    MODE_BGR_32,
    MODE_RGB_24,
    MODE_BGR_24,
    MODE_RGB_16,
    MODE_BGR_16,
    MODE_RGB_15,
    MODE_BGR_15
};

typedef struct xoutput_s xoutput_t;
struct xoutput_s
{
    Display *display;
    int screen;
    Window root;
    Visual *visual;
    unsigned int defaultdepth;
    unsigned long blackpixel;
    int width;
    int height;
    int depth;
    Window wnd;
    XImage *xi;
    XShmSegmentInfo shareme;
    GC gc;
    int shmcompletion;
    unsigned char *framebuffer;
    unsigned char *writeptr;
};

static Pixmap mypixmap;

xoutput_t *xoutput_new( int width, int height )
{
    xoutput_t *output = (xoutput_t *) malloc( sizeof( xoutput_t ) );
    char *window_name = "Text crawl viewer";
    char *icon_name = "Text crawl viewer";
    XSizeHints size_hints;
    XWindowAttributes attribs;

    if( !output ) {
        return 0;
    }

    if( !vsync_init() ) {
        free( output );
        return 0;
    }

    output->display = XOpenDisplay( 0 );
    if( !output->display ) {
        fprintf( stderr, "output: Can't open display '%s'.\n", getenv( "DISPLAY" ) ? getenv( "DISPLAY" ) : "" );
        free( output );
        return 0;
    }
    output->screen = DefaultScreen( output->display );
    output->root = XRootWindow( output->display, output->screen );
    output->blackpixel = BlackPixel( output->display, output->screen );
    output->visual = DefaultVisual( output->display, output->screen );
    output->defaultdepth = DefaultDepth( output->display, output->screen );
    output->width = width;
    output->height = height;

    size_hints.flags = PPosition | PSize | PMinSize | PMaxSize;
    size_hints.max_width = width;
    size_hints.min_width = width;
    size_hints.max_height = height;
    size_hints.min_height = height;

    // Create the window
    output->wnd = XCreateSimpleWindow( output->display, output->root,
                                       0, 0, width, height, 0,
                                       output->blackpixel, output->blackpixel );

    // Set input stuff.
    XSelectInput( output->display, output->wnd, ButtonPressMask | StructureNotifyMask |
                                                KeyPressMask | ExposureMask | PropertyChangeMask );

    // Set the properties
    XSetStandardProperties( output->display, output->wnd, window_name, icon_name,
        0, 0, 0, &size_hints );

    // Map it on the screen
    XMapWindow( output->display, output->wnd );

    // Create a GC
    output->gc = XCreateGC( output->display, output->wnd, 0, 0 );

    output->shareme.shmid = shmget( IPC_PRIVATE, width * height * 4, IPC_CREAT | 0700 );
    output->shareme.shmaddr = (char *) shmat( output->shareme.shmid, 0, 0 );
    output->shareme.readOnly = False;

    XShmAttach( output->display, &output->shareme );
    XSync( output->display, False );
    shmctl( output->shareme.shmid, IPC_RMID, 0 );

    CompletionType = XShmGetEventBase(output->display) + ShmCompletion;

    XGetWindowAttributes( output->display, output->wnd, &attribs );
    output->depth = attribs.depth;
 
    output->xi = XShmCreateImage( output->display, output->visual, output->depth,
                                  ZPixmap, output->shareme.shmaddr, &output->shareme, width, height );
    output->writeptr = (unsigned char *) output->xi->data;
    fprintf( stderr, "xoutput: At depth %d, padded to %d bits per pixel.\n",
             output->depth, output->xi->bits_per_pixel );

    if( output->xi->bits_per_pixel != 32 ) {
        output->framebuffer = (unsigned char *) malloc( width * height * 4 );
        if( !output->framebuffer ) {
            fprintf( stderr, "you suck mcbain.\n" );
        }
    } else {
        output->framebuffer = output->writeptr;
    }

    XInternAtoms( output->display, atomNames, 2, False, &wmProtocolsAtom );
    /* collaborate with the window manager for close requests */
    XSetWMProtocols( output->display, output->wnd, &wmDeleteAtom, 1 );

    mypixmap = XCreatePixmap( output->display, output->wnd, width, height, 24 );

    return output;
}

void xoutput_delete( xoutput_t *output )
{
    XFreeGC( output->display, output->gc );
    XShmDetach( output->display, &output->shareme );
    XDestroyImage( output->xi );
    shmdt( (char *) output->writeptr );
    vsync_shutdown();
}

unsigned char *xoutput_get_buffer( xoutput_t *output )
{
    return output->framebuffer;
}

int xoutput_get_stride( xoutput_t *output )
{
    return (output->width * 4);
}

static Bool predicate(Display *dpy, XEvent *ev, XPointer arg)
{
  if(ev->type == CompletionType) {
    return True;
  } else {
    return False;
  }
}

void xoutput_wait( xoutput_t *output )
{
    vsync_wait_for_retrace();
}

void xoutput_send_next_frame( xoutput_t *output )
{
    XEvent ev;

    if( output->depth == 16 ) {
        rgb32_to_rgb16_scanline( output->writeptr, output->framebuffer,
                                 output->width * output->height );
    }

    if( use_xshm_directly ) {
        XShmPutImage( output->display, output->wnd,
                      output->gc, output->xi, 0, 0, 0, 0,
                      output->width, output->height, True );
        XSync( output->display, False );

        /**
         * This is to make sure that we are free to use the image again
         * It waits for an XShmCompletionEvent
         */
        XIfEvent( output->display, &ev, predicate, 0 );
        // XSync( output->display, False );
    } else {
        XLockDisplay( output->display );
        XShmPutImage( output->display, mypixmap,
                      output->gc, output->xi, 0, 0, 0, 0,
                      output->width, output->height, True );
        XUnlockDisplay( output->display );
        XFlush( output->display );
        XSync( output->display, False );

        /**
         * This is to make sure that we are free to use the image again
         * It waits for an XShmCompletionEvent
         */
        XIfEvent( output->display, &ev, predicate, 0 );
        XCopyArea( output->display, mypixmap, output->wnd, output->gc,
                   0, 0, output->width, output->height, 0, 0 );
        XSync( output->display, False );
    }
}

void xoutput_handle_events( xoutput_t *output, input_t *in )
{
    XEvent event;
    KeySym mykey;
    int arg = 0;
    char mykey_string;
    int reconfwidth = 0;
    int reconfheight = 0;

    while( XPending( output->display ) ) {
        XNextEvent( output->display, &event );

        switch( event.type ) {
        case ClientMessage:
            /* X sucks this way.  Thanks to walken for help on this one. */
            if( ( event.xclient.message_type == wmProtocolsAtom ) && ( event.xclient.format == 32 ) &&
                ( event.xclient.data.l[ 0 ] == wmDeleteAtom ) ) {
                input_callback( in, I_QUIT, 0 );
            }
            break;
        case DestroyNotify:
            if( event.xdestroywindow.window != output->wnd ) {
                break;
            }
            input_callback( in, I_QUIT, 0 );
            break;
        case Expose:
            break;
        case ConfigureNotify:
            reconfwidth = event.xconfigure.width;
            reconfheight = event.xconfigure.height;
            break;
        case KeyPress:
            mykey = XKeycodeToKeysym( output->display, event.xkey.keycode, 0 );
            if( event.xkey.state & ShiftMask ) arg |= I_SHIFT;
            if( event.xkey.state & ControlMask ) arg |= I_CTRL;
            if( event.xkey.state & Mod1Mask ) arg |= I_META;

            if( mykey >= XK_a && mykey <= XK_z ) {
                arg |= mykey - XK_a + 'a';
            } else if( mykey >= XK_A && mykey <= XK_Z ) {
                arg |= mykey - XK_A + 'a';
            } else if( mykey >= XK_0 && mykey <= XK_9 ) {
                arg |= mykey - XK_0 + '0';
            } else {
                XLookupString( &event.xkey, &mykey_string, 1, 0, 0 );
                arg |= mykey_string;
            }

            switch( mykey ) {
            case XK_Escape: arg |= I_ESCAPE; break;

            case XK_Home:      arg |= I_HOME; break;
            case XK_End:       arg |= I_END; break;
            case XK_Left:      arg |= I_LEFT; break;
            case XK_Up:        arg |= I_UP; break;
            case XK_Right:     arg |= I_RIGHT; break;
            case XK_Down:      arg |= I_DOWN; break;
            case XK_Page_Up:   arg |= I_PGUP; break;
            case XK_Page_Down: arg |= I_PGDN; break;

            case XK_F1: arg |= I_F1; break;
            case XK_F2: arg |= I_F2; break;
            case XK_F3: arg |= I_F3; break;
            case XK_F4: arg |= I_F4; break;
            case XK_F5: arg |= I_F5; break;
            case XK_F6: arg |= I_F6; break;
            case XK_F7: arg |= I_F7; break;
            case XK_F8: arg |= I_F8; break;
            case XK_F9: arg |= I_F9; break;
            case XK_F10: arg |= I_F10; break;
            case XK_F11: arg |= I_F11; break;
            case XK_F12: arg |= I_F12; break;

            case XK_KP_Enter: arg |= I_ENTER; break;
            case XK_KP_0: arg |= '0'; break;
            case XK_KP_1: arg |= '1'; break;
            case XK_KP_2: arg |= '2'; break;
            case XK_KP_3: arg |= '3'; break;
            case XK_KP_4: arg |= '4'; break;
            case XK_KP_5: arg |= '5'; break;
            case XK_KP_6: arg |= '6'; break;
            case XK_KP_7: arg |= '7'; break;
            case XK_KP_8: arg |= '8'; break;
            case XK_KP_9: arg |= '9'; break;
            case XK_KP_Multiply: arg |= '*'; break;
            case XK_KP_Subtract: arg |= '-'; break;
            case XK_KP_Add: arg |= '+'; break;
            case XK_KP_Divide: arg |= '/'; break;

            case XK_KP_End: arg |= '1'; break;
            case XK_KP_Down: arg |= '2'; break;
            case XK_KP_Next: arg |= '3'; break;
            case XK_KP_Left: arg |= '4'; break;
            case XK_KP_Begin: arg |= '5'; break;
            case XK_KP_Right: arg |= '6'; break;
            case XK_KP_Home: arg |= '7'; break;
            case XK_KP_Up: arg |= '8'; break;
            case XK_KP_Prior: arg |= '9'; break;

            default: break;
            }
            input_callback( in, I_KEYDOWN, arg );
            break;
        case ButtonPress:
            input_callback( in, I_BUTTONPRESS, event.xbutton.button );
            break;
        default: break;
        }
    }
}

/**
 * Abstract API implementation.
 */
static xoutput_t *output = 0;

int xoutput_init( int width, int height )
{
    output = xoutput_new( width, height );
    if( !output ) return 0; else return 1;
}

void xoutput_shutdown( void )
{
    xoutput_delete( output );
}

void xoutput_poll_events( input_t *in )
{
    xoutput_handle_events( output, in );
}

void xoutput_lock_output_buffer( void )
{
}

void xoutput_wait_for_retrace( void )
{
    xoutput_wait( output );
}

unsigned char *xoutput_get_output_buffer( void )
{
    return xoutput_get_buffer( output );
}

int xoutput_get_output_stride( void )
{
    return xoutput_get_stride( output );
}

void xoutput_unlock_output_buffer( void )
{
}

void xoutput_set_window_caption( const char *caption )
{
}

void xoutput_show_frame( void )
{
    xoutput_send_next_frame( output );
}

output_api_t xoutput_api = {
    xoutput_init,
    xoutput_shutdown,

    xoutput_lock_output_buffer,
    xoutput_get_output_buffer,
    xoutput_get_output_stride,
    xoutput_unlock_output_buffer,

    xoutput_show_frame,
    xoutput_wait_for_retrace,
    xoutput_set_window_caption,
    xoutput_poll_events
};

output_api_t *xoutput_get_output( void )
{
    return &xoutput_api;
}

